import json
from dataclasses import dataclass, field
from typing import List, Type, Self, override

from busline.event.message.number_message import Int64Message, Int32Message, Float64Message, Float32Message
from busline.event.message.string_message import StringMessage
from dataclasses_avroschema import AvroModel

from busline.event.message.avro_message import AvroMessageMixin


@dataclass(kw_only=True)
class SchemaSpec(AvroModel):
    """

    Author: Nicola Ricciardi
    """

    schemas: List[str] = field(default_factory=list)
    support_empty_schema: bool = field(default=False)
    support_undefined_schema: bool = field(default=False)

    def with_empty_support(self) -> Self:
        self.support_empty_schema = True
        return self

    @property
    def has_some_explicit_schemas(self) -> bool:
        return len(self.schemas) > 0

    @classmethod
    def from_schema(cls, schema: str) -> Self:
        return cls(schemas=[schema])

    @classmethod
    def empty(cls) -> Self:
        return cls(support_empty_schema=True)

    @classmethod
    def undefined(cls) -> Self:
        return cls(schemas=[], support_empty_schema=False)

    @classmethod
    def from_message(cls, payload: Type[AvroMessageMixin]) -> Self:
        return cls.from_schema(payload.avro_schema())

    @classmethod
    def int64(cls) -> Self:
        return cls.from_message(Int64Message)

    @classmethod
    def int32(cls) -> Self:
        return cls.from_message(Int32Message)

    @classmethod
    def float64(cls) -> Self:
        return cls.from_message(Float64Message)

    @classmethod
    def float32(cls) -> Self:
        return cls.from_message(Float32Message)

    @classmethod
    def string(cls) -> Self:
        return cls.from_message(StringMessage)


    def is_compatible(self, other: Self) -> bool:
        if self.support_undefined_schema != other.support_undefined_schema:
            return False

        if self.support_empty_schema != other.support_empty_schema:
            return False

        if len(self.schemas) != len(other.schemas):
            return False

        for my_schema in self.schemas:
            found = False

            for other_schema in other.schemas:
                if self._compare_two_schema(my_schema, other_schema):
                    found = True
                    break

            if not found:
                return False

        return True

    def is_compatible_with_schema(self, target_schema: str) -> bool:
        if self.support_undefined_schema:
            return True

        for my_schema in self.schemas:
            if self._compare_two_schema(my_schema, target_schema):
                return True

        return False


    @classmethod
    def _compare_two_schema(cls, schema_a: str, schema_b: str):
        """
        Compare two schemas and return True if they are equal
        """

        try:
            schema_a_dict = json.loads(schema_a)
            schema_b_dict = json.loads(schema_b)

            return schema_a_dict == schema_b_dict
        except:
            return schema_a == schema_b


@dataclass
class Input(SchemaSpec):

    @property
    def has_input(self) -> Self:
        return self.support_undefined_schema or self.support_empty_schema or self.has_some_explicit_schemas

    @classmethod
    def no_input(cls) -> Self:
        return cls()

    @override
    def is_compatible(self, other: Self) -> bool:
        if int(other.has_input) + int(self.has_input) == 1:
            return False

        return super().is_compatible(other)

@dataclass(kw_only=True)
class Output(SchemaSpec):

    @property
    def has_output(self) -> Self:
        return self.support_undefined_schema or self.support_empty_schema or self.has_some_explicit_schemas

    @classmethod
    def no_output(cls) -> Self:
        return cls()

    @override
    def is_compatible(self, other: Self) -> bool:
        if int(other.has_output) + int(self.has_output) == 1:
            return False

        return super().is_compatible(other)


@dataclass(kw_only=True)
class Inputs(AvroModel):
    inputs: List[Input]

    def input_is_compatible(self, input: Input) -> bool:
        found = False
        for my_input in self.inputs:

            if my_input.is_compatible(input):
                found = True
                break

        return found


@dataclass(kw_only=True)
class Outputs(AvroModel):
    outputs: List[Output]

    def output_is_compatible(self, output: Output) -> bool:
        found = False
        for my_output in self.outputs:

            if my_output.is_compatible(output):
                found = True
                break

        return found




from dataclasses import dataclass
from typing import Optional

from busline.event.registry import add_to_registry
from busline.event.message.avro_message import AvroMessageMixin


@dataclass(frozen=True, kw_only=True)
class ConfirmConnectionMessage(AvroMessageMixin):
    """
    Plugin --- confirm connection ---> Core

    Message used by plugins to confirm the connection creation

    Author: Nicola Ricciardi
    """

    plugin_identifier: str
    operation_name: str
    operation_input_topic: Optional[str]
    plugin_side_close_operation_connection_topic: str


@dataclass(frozen=True, kw_only=True)
class OperationNoLongerAvailableMessage(AvroMessageMixin):
    """
    Plugin --- operation no longer available ---> Core

    Message used by plugins to notify core that operation is no longer available

    Author: Nicola Ricciardi
    """

    plugin_identifier: str
    operation_name: str


import inspect
from dataclasses import dataclass, field
from typing import Dict, Self, Any

from busline.client.subscriber.event_handler import event_handler
from busline.client.subscriber.event_handler.event_handler import EventHandler


@dataclass(kw_only=True)
class _SinkDescriptor:
    operation_name: str
    func: Any

    def __post_init__(self):
        self.func = event_handler(self.func)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.operation_name not in instance.operation_sinks:
            instance.operation_sinks[self.operation_name] = self.func.__get__(instance, owner)

        return self.func.__get__(instance, owner)


def sink(operation_name: str):

    def decorator(func):
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Event handler must be async")

        return _SinkDescriptor(
            func=func,
            operation_name=operation_name
        )

    return decorator




@dataclass(kw_only=True)
class SinksProviderMixin:

    operation_sinks: Dict[str, EventHandler] = field(default_factory=dict, init=False)    # operation_name => EventHandler

    def __post_init__(self):
        # used to refresh sinks
        for attr_name in dir(self):
            _ = getattr(self, attr_name)

    def with_operation_sink(self, operation_name: str, handler: EventHandler) -> Self:

        self.operation_sinks[operation_name] = handler

        return self



import asyncio
import logging
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Set
import uuid

from busline.client.pubsub_client import PubSubClient
from busline.client.subscriber.event_handler import event_handler
from busline.event.event import Event
from orbitalis.events.close_connection import GracefulCloseConnectionMessage, GracelessCloneConnectionMessage, \
    CloseConnectionAckMessage
from orbitalis.events.keepalive import KeepaliveRequestMessage, KeepaliveMessage
from orbitalis.orbiter.connection import Connection
from orbitalis.orbiter.pending_request import PendingRequest
from orbitalis.orbiter.schemaspec import Output, Input
from orbitalis.plugin.operation import Operation

DEFAULT_DISCOVER_TOPIC = "$handshake.discover"
DEFAULT_LOOP_INTERVAL = 1
DEFAULT_PENDING_REQUESTS_EXPIRE_AFTER = 60.0
DEFAULT_SEND_KEEPALIVE_BEFORE_TIMELIMIT = 10.0
DEFAULT_CONSIDERED_DEAD_AFTER = 120.0
DEFAULT_GRACEFUL_CLOSE_TIMEOUT = 300.0


@dataclass(kw_only=True)
class Orbiter(ABC):
    """
    Base class which provides common capabilities to components.
    It manages pending requests, connections, keepalive and connection close procedure.
    In addiction, it has useful shared methods and main loop.

    Author: Nicola Ricciardi
    """

    eventbus_client: PubSubClient

    identifier: str = field(default_factory=lambda: str(uuid.uuid4()))

    discover_topic: str = field(default=DEFAULT_DISCOVER_TOPIC)
    raise_exceptions: bool = field(default=False)

    loop_interval: float = field(default=DEFAULT_LOOP_INTERVAL)

    close_connection_if_unused_after: Optional[float] = field(default=None)
    pending_requests_expire_after: Optional[float] = field(default=DEFAULT_PENDING_REQUESTS_EXPIRE_AFTER)
    consider_others_dead_after: Optional[float] = field(default=DEFAULT_CONSIDERED_DEAD_AFTER)
    send_keepalive_before_timelimit: float = field(default=DEFAULT_SEND_KEEPALIVE_BEFORE_TIMELIMIT)
    graceful_close_timeout: Optional[float] = field(default=DEFAULT_GRACEFUL_CLOSE_TIMEOUT)

    with_loop: bool = field(default=True)

    new_connection_added_event: asyncio.Event = field(default_factory=asyncio.Event, init=False)

    _others_considers_me_dead_after: Dict[str, float] = field(default_factory=dict, init=False)     # remote_identifier => time
    _remote_keepalive_request_topics: Dict[str, str] = field(default_factory=dict, init=False)   # remote_identifier => keepalive_request_topic
    _remote_keepalive_topics: Dict[str, str] = field(default_factory=dict, init=False)   # remote_identifier => keepalive_topic
    _last_seen: Dict[str, datetime] = field(default_factory=dict, init=False)   # remote_identifier => datetime
    _last_keepalive_sent: Dict[str, datetime] = field(default_factory=dict, init=False)   # remote_identifier => datetime

    _connections: Dict[str, Dict[str, Connection]] = field(default_factory=lambda: defaultdict(dict), init=False)    # remote_identifier => { operation_name => Connection }
    _pending_requests: Dict[str, Dict[str, PendingRequest]] = field(default_factory=lambda: defaultdict(dict), init=False)    # remote_identifier => { operation_name => PendingRequest }

    _unsubscribe_on_full_close_bucket: Dict[str, Set[str]] = field(default_factory=lambda: defaultdict(set), init=False)

    __stop_loop_controller: asyncio.Event = field(default_factory=lambda: asyncio.Event(), init=False)
    __pause_loop_controller: asyncio.Event = field(default_factory=lambda: asyncio.Event(), init=False)

    def __post_init__(self):
        if 0 > self.send_keepalive_before_timelimit:
            raise ValueError("send_keepalive_before_timelimit must be >= 0")
        
        self.new_connection_added_event.clear()

    @property
    def keepalive_request_topic(self) -> str:
        return f"$keepalive.{self.identifier}.request"

    @property
    def keepalive_topic(self) -> str:
        return f"$keepalive.{self.identifier}"

    @property
    def _all_pending_requests(self) -> List[PendingRequest]:

        pending_requests: List[PendingRequest] = []
        for of_operation in self._pending_requests.values():
            for pending_request in of_operation.values():
                pending_requests.append(pending_request)

        return pending_requests

    @property
    def _all_connections(self) -> List[Connection]:

        connections: List[Connection] = []
        for of_operation in self._connections.values():
            for connection in of_operation.values():
                connections.append(connection)

        return connections

    @property
    def dead_remote_identifiers(self) -> List[str]:
        if self.consider_others_dead_after is None:
            return []

        dead = []
        now = datetime.now()
        for remote_identifier, last_seen in self._last_seen.items():
            if (last_seen + timedelta(seconds=self.consider_others_dead_after)) < now:
                dead.append(remote_identifier)

        return dead
    
    @property
    def remote_identifiers(self) -> Set[str]:
        identifiers: Set[str] = set()

        for connection in self._all_connections:
            identifiers.add(connection.remote_identifier)

        return identifiers


    async def _get_on_close_data(self, remote_identifier: str, operation_name: str) -> Optional[bytes]:
        """
        Hook used to obtain data to send on close connection, by default None is returned
        """

        return None


    async def start(self, *args, **kwargs):
        logging.info("%s: starting...", self)
        await self._on_starting(*args, **kwargs)
        await self._internal_start(*args, **kwargs)
        await self._on_started(*args, **kwargs)
        logging.info("%s: started", self)

    async def _on_starting(self, *args, **kwargs):
        """
        Hook called before starting
        """

    async def _internal_start(self, *args, **kwargs):
        """
        Actual implementation to start the orbiter
        """

        await self.eventbus_client.connect()

        await asyncio.gather(
            self.eventbus_client.subscribe(
                self.keepalive_request_topic,
                self.__keepalive_request_event_handler
            ),
            self.eventbus_client.subscribe(
                self.keepalive_topic,
                self.__keepalive_event_handler
            )
        )

        if self.with_loop:
            self.start_loop()

    async def _on_started(self, *args, **kwargs):
        """
        Hook called after starting
        """

    async def stop(self, *args, **kwargs):
        logging.info("%s: stopping...", self)
        await self._on_stopping(*args, **kwargs)
        await self._internal_stop(*args, **kwargs)
        await self._on_stopped(*args, **kwargs)
        logging.info("%s: stopped", self)

    async def _on_stopping(self, *args, **kwargs):
        """
        Hook called before stopping
        """

    async def _internal_stop(self, *args, **kwargs):
        """
        Actual implementation to stop the orbiter
        """

        tasks = []

        tasks.append(
            asyncio.create_task(
                self.eventbus_client.multi_unsubscribe([
                    self.keepalive_request_topic,
                    self.keepalive_topic,
                ])
            )
        )

        for connection in self._all_connections:
            tasks.append(
                asyncio.create_task(
                    self.send_graceless_close_connection(
                        remote_identifier=connection.remote_identifier,
                        operation_name=connection.operation_name,
                        data=await self._get_on_close_data(connection.remote_identifier, connection.operation_name)
                    )
                )
            )

        self.stop_loop()

        await asyncio.gather(*tasks)

    async def _on_stopped(self, *args, **kwargs):
        """
        Hook called after stopping
        """

    def _connections_by_remote_identifier(self, remote_identifier: str) -> Dict[str, Connection]:
        return self._connections[remote_identifier]

    def _add_connection(self, connection: Connection):
        self.new_connection_added_event.clear()
        self._connections[connection.remote_identifier][connection.operation_name] = connection
        self.new_connection_added_event.set()

    def _remove_connection(self, connection: Connection) -> Optional[Connection]:
        if connection.remote_identifier in self._connections:
            if connection.operation_name in self._connections[connection.remote_identifier]:
                return self._connections[connection.remote_identifier].pop(connection.operation_name)

        raise ValueError(f"{self}: no connection for identifier '{connection.remote_identifier}' and operation '{connection.operation_name}'")

    def _pending_requests_by_remote_identifier(self, remote_identifier: str) -> Dict[str, PendingRequest]:
        return self._pending_requests[remote_identifier]

    def _is_pending(self, remote_identifier: str, operation_name: str) -> bool:
        if remote_identifier in self._pending_requests:
            if operation_name in self._pending_requests[remote_identifier]:
                return True

        return False

    def _add_pending_request(self, pending_request: PendingRequest):
        self._pending_requests[pending_request.remote_identifier][pending_request.operation_name] = pending_request

    def _remove_pending_request(self, pending_request: PendingRequest) -> Optional[PendingRequest]:
        if pending_request.remote_identifier in self._pending_requests:
            if pending_request.operation_name in self._pending_requests[pending_request.remote_identifier]:
                return self._pending_requests[pending_request.remote_identifier].pop(pending_request.operation_name)

        raise ValueError(f"{self}: no pending request for identifier '{pending_request.remote_identifier}' and operation '{pending_request.operation_name}'")


    def retrieve_connections(self, *, remote_identifier: Optional[str] = None, input_topic: Optional[str] = None,
                             output_topic: Optional[str] = None, operation_name: Optional[str] = None,
                             input: Optional[Input] = None, output: Optional[Output] = None) -> List[Connection]:
        """
        Retrieve all connections which satisfy query
        """

        connections: List[Connection] = []

        for remote_identifier_, operation_name_connection in self._connections.items():
            if remote_identifier is not None and remote_identifier != remote_identifier_:
                continue

            for operation_name_, connection in operation_name_connection.items():
                assert operation_name_ == connection.operation_name

                if operation_name is not None and operation_name != operation_name_:
                    continue

                if input_topic is not None and input_topic != connection.input_topic:
                    continue

                if output_topic is not None and output_topic != connection.output_topic:
                    continue

                if input is not None and not input.is_compatible(connection.input):
                    continue

                if output is not None and not output.is_compatible(connection.output):
                    continue

                connections.append(connection)

        return connections

    def _find_connection_or_fail(self, input_topic: str, operation_name: str) -> Connection:
        """
        Find the connection associated to given input topic and operation name.
        Raise ValueError if too many connections are found or if there are no connections.
        """

        connections = self.retrieve_connections(
            input_topic=input_topic,
            operation_name=operation_name
        )

        if len(connections) <= 0:
            raise ValueError("Connection not found")

        if len(connections) >= 2:
            raise ValueError("Too many connections found")

        return connections[0]

    async def retrieve_and_touch_connections(self, operation_name: str, *, input_topic: Optional[str] = None) -> List[Connection]:
        """
        Retrieve connections based on operation's name and input topic, then *lock* and touch connections.
        Finally, connections are returned.
        """

        connections = self.retrieve_connections(
            input_topic=input_topic,
            operation_name=operation_name
        )

        for connection in connections:
            async with connection.lock:
                connection.touch()

        return connections

    def _on_promote_pending_request_to_connection(self, pending_request: PendingRequest):
        """
        Hook called before promotion
        """

    def _promote_pending_request_to_connection(self, pending_request: PendingRequest):
        """
        Transform a pending request into a connection
        """

        try:

            self._on_promote_pending_request_to_connection(pending_request)

            self._add_connection(pending_request.into_connection())
            self._remove_pending_request(pending_request)

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

    async def discard_expired_pending_requests(self) -> int:
        """
        Remove expired pending requests.
        Return total amount of discarded requests
        """

        if self.pending_requests_expire_after is None:
            return 0

        if len(self._pending_requests) == 0:
            return 0

        to_remove: List[PendingRequest] = []

        discarded = 0
        for remote_identifier, of_operation in self._pending_requests.items():
            for operation_name, pending_request in of_operation.items():
                if (pending_request.created_at + timedelta(seconds=self.pending_requests_expire_after)) < datetime.now():
                    to_remove.append(pending_request)

        for pending_request in to_remove:
            async with pending_request.lock:
                try:
                    self._remove_pending_request(pending_request)
                    discarded += 1
                except Exception as e:
                    logging.warning("%s: pending request %s was removed before discarding", self, pending_request)

        return discarded

    async def close_unused_connections(self) -> int:
        """
        Send a graceful close request to all remote orbiter if connection was unused based on close_connection_if_unused_after
        """

        try:
            if self.close_connection_if_unused_after is None:
                return 0

            if len(self._connections) == 0:
                return 0

            to_close: List[Connection] = []

            closed = 0
            for remote_identifier, of_operation in self._connections.items():
                for operation_name, connection in of_operation.items():
                    expiration= connection.created_at + timedelta(seconds=self.close_connection_if_unused_after)
                    if expiration < datetime.now():
                        to_close.append(connection)

            tasks = []
            for connection in to_close:
                tasks.append(
                    asyncio.create_task(
                        self.send_graceful_close_connection(connection.remote_identifier, connection.operation_name)
                    )
                )
                closed += 1

            if len(tasks) > 0:
                await asyncio.gather(*tasks)

            return closed

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

        return 0

    async def force_close_connection_for_out_to_timeout_pending_graceful_close_connection(self) -> int:
        try:
            if self.graceful_close_timeout is None:
                return 0

            if len(self._connections) == 0:
                return 0

            to_close: List[Connection] = []

            closed = 0
            for remote_identifier, of_operation in self._connections.items():
                for operation_name, connection in of_operation.items():
                    if connection.soft_closed_at is None:
                        continue

                    expiration= connection.soft_closed_at + timedelta(seconds=self.graceful_close_timeout)
                    if expiration < datetime.now():
                        to_close.append(connection)

            tasks = []
            for connection in to_close:
                tasks.append(
                    asyncio.create_task(
                        self.send_graceless_close_connection(connection.remote_identifier, connection.operation_name)
                    )
                )
                closed += 1

            if len(tasks) > 0:
                await asyncio.gather(*tasks)

            return closed

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

        return 0

    def update_acquaintances(self, remote_identifier: str,
        *, keepalive_topic: str, keepalive_request_topic: str, consider_me_dead_after: float):
        """
        Update knowledge about keepalive request topics, keepalive topics and dead time
        """

        self._remote_keepalive_request_topics[remote_identifier] = keepalive_request_topic

        self._remote_keepalive_topics[remote_identifier] = keepalive_topic

        self._others_considers_me_dead_after[remote_identifier] = consider_me_dead_after

    def have_seen(self, remote_identifier: str, *, when: Optional[datetime] = None):
        """
        Update last seen for remote orbiter
        """

        if when is None:
            when = datetime.now()

        self._last_seen[remote_identifier] = when

    def clear_last_seen(self):
        self._last_seen = dict()

    async def _on_keepalive_request(self, from_identifier: str):
        """
        Hook called on keepalive request, before response
        """

    @event_handler
    async def __keepalive_request_event_handler(self, topic: str, event: Event[KeepaliveRequestMessage]):
        try:
            await self._on_keepalive_request(event.payload.from_identifier)

            await self.eventbus_client.publish(
                event.payload.keepalive_topic,
                KeepaliveMessage(from_identifier=self.identifier)
            )
        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

    async def _on_keepalive(self, from_identifier: str):
        """
        Hook called on inbound keepalive
        """

    @event_handler
    async def __keepalive_event_handler(self, topic: str, event: Event[KeepaliveMessage]):
        await self._on_keepalive(event.payload.from_identifier)

        self._last_seen[event.payload.from_identifier] = datetime.now()

    async def send_keepalive(self, remote_identifier: str):

        if remote_identifier not in self._remote_keepalive_topics:
            raise ValueError(f"Keepalive topic not found for {remote_identifier}")

        await self.eventbus_client.publish(
            self._remote_keepalive_topics[remote_identifier],
            KeepaliveMessage(
                from_identifier=self.identifier
            )
        )

        self._last_keepalive_sent[remote_identifier] = datetime.now()

    async def send_keepalive_request(self, *, keepalive_request_topic: Optional[str] = None, remote_identifier: Optional[str] = None):

        if keepalive_request_topic is None and remote_identifier is None:
            raise ValueError("Missed target")

        if remote_identifier is not None:
            keepalive_request_topic = self._remote_keepalive_request_topics[remote_identifier]

        await self.eventbus_client.publish(
            keepalive_request_topic,
            KeepaliveRequestMessage(
                from_identifier=self.identifier,
                keepalive_topic=self.keepalive_topic
            )
        )

    async def send_all_keepalive_based_on_connections(self):
        """
        Send keepalive messages to all remote orbiters which have a connection with this orbiter
        """

        tasks = []
        for remote_identifier, operations in self._connections.values():
            if len(operations) > 0 and remote_identifier in self._remote_keepalive_topics:
                tasks.append(
                    asyncio.create_task(self.send_keepalive(remote_identifier=remote_identifier))
                )

        await asyncio.gather(*tasks)

    async def send_keepalive_based_on_connections_and_threshold(self):
        """
        Send keepalive messages to all remote orbiters which have a connection with this orbiter only if
        send_keepalive_before_timelimit seconds away from being considered dead this orbiter
        """

        tasks = []

        for remote_identifier in self._connections.keys():
            if remote_identifier not in self._others_considers_me_dead_after:
                logging.error("%s: no dead time associated to %s, keepalive sending skipped", self, remote_identifier)
                continue

            if remote_identifier not in self._last_keepalive_sent:
                logging.debug("%s: not previous keepalive sent to %s", self, remote_identifier)
                tasks.append(
                    asyncio.create_task(
                        self.send_keepalive(remote_identifier)
                    )
                )
                continue

            considered_dead_at: datetime = self._last_keepalive_sent[remote_identifier] + timedelta(
                seconds=self._others_considers_me_dead_after[remote_identifier])

            if (considered_dead_at - datetime.now()).seconds < 0:
                logging.warning("%s: %s could be flag me as dead, anyway keepalive will be sent", self, remote_identifier)

            assert 0 <= self.send_keepalive_before_timelimit, "send_keepalive_threshold_multiplier must be >= 0"

            if remote_identifier not in self._last_keepalive_sent \
                    or (considered_dead_at - datetime.now()).seconds < self.send_keepalive_before_timelimit:
                tasks.append(
                    asyncio.create_task(
                        self.send_keepalive(remote_identifier)
                    )
                )

        await asyncio.gather(*tasks)


    def _build_incoming_close_connection_topic(self, remote_identifier: str, operation_name: str) -> str:
        return f"{operation_name}.{self.identifier}.{remote_identifier}.close"

    async def send_graceless_close_connection(self, remote_identifier: str, operation_name: str, data: Optional[bytes] = None):
        """
        Send a graceless close connection request to specified remote orbiter. Therefore, self side connection will be closed immediately
        """

        try:

            await self._on_graceless_close_connection(remote_identifier, operation_name, data)

            connection = await self._close_self_side_connection(
                remote_identifier,
                operation_name
            )

            await self.eventbus_client.publish(
                connection.close_connection_to_remote_topic,
                GracelessCloneConnectionMessage(
                    from_identifier=self.identifier,
                    operation_name=operation_name,
                    data=data
                )
            )

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

    async def _on_graceless_close_connection(self, remote_identifier: str, operation_name: str, data: Optional[bytes]):
        """
        Hook called before graceless close connection request is sent
        """

    async def _on_close_connection(self, connection: Connection):
        """
        Hook called when a connection is closed
        """

    async def _close_self_side_connection(self, remote_identifier: str, operation_name: str):
        """
        Close local connection with remote orbiter, therefore only this orbiter will no longer be able to use connection.
        Generally, a close connection request was sent before this method call.
        """

        connection = self._connections[remote_identifier][operation_name]

        async with connection.lock:
            connection = self._remove_connection(connection)

        await self._on_close_connection(connection)

        close_incoming_close_connection_task = self.eventbus_client.unsubscribe(connection.incoming_close_connection_topic)

        if len(self._connections[remote_identifier].values()) == 0:
            await self.eventbus_client.multi_unsubscribe(list(self._unsubscribe_on_full_close_bucket[remote_identifier]), parallelize=True)
            self._unsubscribe_on_full_close_bucket.pop(remote_identifier, None)

        await close_incoming_close_connection_task

        logging.info("%s: self side connection %s closed", self, connection)

        return connection

    def _build_ack_close_topic(self, remote_identifier: str, operation_name: str) -> str:
        return f"{operation_name}.{self.identifier}.{remote_identifier}.close.ack"

    async def send_graceful_close_connection(self, remote_identifier: str, operation_name: str, data: Optional[bytes] = None):
        """
        Send a graceful close connection request to specified remote orbiter, therefore self side connection is not close immediately, but ACK is waited
        """

        try:
            await self._on_graceful_close_connection(remote_identifier, operation_name, data)

            connections = self.retrieve_connections(
                remote_identifier=remote_identifier,
                operation_name=operation_name
            )

            assert len(connections) == 1

            connection = connections[0]

            close_connection_to_remote_topic = None
            async with connection.lock:
                connection.soft_close()
                close_connection_to_remote_topic = connection.close_connection_to_remote_topic

            ack_topic = self._build_ack_close_topic(remote_identifier, operation_name)

            await self.eventbus_client.subscribe(ack_topic, self.__close_connection_ack_event_handler)

            assert close_connection_to_remote_topic is not None

            await self.eventbus_client.publish(
                close_connection_to_remote_topic,
                GracefulCloseConnectionMessage(
                    from_identifier=self.identifier,
                    operation_name=operation_name,
                    ack_topic=ack_topic,
                    data=data
                )
            )

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

    async def _on_graceful_close_connection(self, remote_identifier: str, operation_name: str, data: Optional[bytes]):
        """
        Hook called before sending graceful close connection request
        """

    async def _graceful_close_connection(self, topic: str, close_connection_message: GracefulCloseConnectionMessage):
        await self._on_graceful_close_connection(
            close_connection_message.from_identifier,
            close_connection_message.operation_name,
            close_connection_message.data
        )

        await self._close_self_side_connection(
            close_connection_message.from_identifier,
            close_connection_message.operation_name
        )

        await self.eventbus_client.publish(
            close_connection_message.ack_topic,
            CloseConnectionAckMessage(
                from_identifier=self.identifier,
                operation_name=close_connection_message.operation_name
            )
        )

    @event_handler
    async def __close_connection_ack_event_handler(self, topic: str, event: Event[CloseConnectionAckMessage]):

        self.have_seen(event.payload.from_identifier)

        await asyncio.gather(
            self._close_self_side_connection(
                event.payload.from_identifier,
                event.payload.operation_name
            ),
            self.eventbus_client.unsubscribe(topic)
        )

    @event_handler
    async def _close_connection_event_handler(self, topic: str, event: Event[GracefulCloseConnectionMessage | GracelessCloneConnectionMessage]):
        try:
            if isinstance(event.payload, GracefulCloseConnectionMessage):
                await self._graceful_close_connection(topic, event.payload)

            elif isinstance(event.payload, GracelessCloneConnectionMessage):
                await self._on_graceless_close_connection(
                    event.payload.from_identifier,
                    event.payload.operation_name,
                    event.payload.data
                )

                await self._close_self_side_connection(
                    event.payload.from_identifier,
                    event.payload.operation_name
                )

            else:
                logging.error("%s: unable to handle close connection event: %s", self, event)

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e



    async def _on_loop_start(self):
        """
        Hook called on loop start
        """

    async def _on_new_loop_iteration(self):
        """
        Hook called before every loop iteration
        """

    async def _on_loop_iteration_end(self):
        """
        Hook called at the end of every loop iteration
        """

    async def _on_loop_iteration(self):
        """
        Hook called during every loop iteration
        """

    def start_loop(self):
        self.__stop_loop_controller.clear()
        self.__pause_loop_controller.clear()
        asyncio.create_task(self.__loop())

    def stop_loop(self):
        self.__stop_loop_controller.set()

    def pause_loop(self):
        self.__pause_loop_controller.set()

    def resume_loop(self):
        self.__pause_loop_controller.clear()

    async def __loop(self):

        await self._on_loop_start()

        while not self.__stop_loop_controller.is_set():

            await asyncio.sleep(self.loop_interval)

            if self.__pause_loop_controller.is_set():
                continue

            try:
                await self._on_new_loop_iteration()

                await asyncio.gather(
                    self._on_loop_iteration(),
                    self.close_unused_connections(),
                    self.discard_expired_pending_requests(),
                    self.force_close_connection_for_out_to_timeout_pending_graceful_close_connection(),
                    self.send_keepalive_based_on_connections_and_threshold(),
                )

                await self._on_loop_iteration_end()

            except Exception as e:

                logging.error("%s: error during loop iteration: %s", self, repr(e))

                if self.raise_exceptions:
                    raise e



from dataclasses import dataclass
from typing import List

from dataclasses_avroschema import AvroModel

from busline.event.registry import add_to_registry
from busline.event.message.avro_message import AvroMessageMixin
from orbitalis.orbiter.schemaspec import Input, Output


@dataclass
class OfferedOperation(AvroModel):
    name: str
    input: Input
    output: Output

@dataclass(frozen=True)
class OfferMessage(AvroMessageMixin):
    """

    Plugin --- offer ---> Core

    Message used by plugins to response to discover message, providing their base information and a list of offered operations.
    List of offered operations can be smaller than fullset provided by discover

    Author: Nicola Ricciardi
    """

    plugin_identifier: str
    offered_operations: List[OfferedOperation]
    reply_topic: str
    considered_dead_after: float
    plugin_keepalive_topic: str
    plugin_keepalive_request_topic: str


import asyncio
import copy
import random
from datetime import datetime, timedelta
import logging
from dataclasses import dataclass, field
from typing import Type, override, Dict, Set, Optional, List
from busline.client.subscriber.event_handler import event_handler
from busline.client.subscriber.event_handler.event_handler import EventHandler
from busline.event.message.avro_message import AvroMessageMixin
from busline.event.event import Event

from orbitalis.core.sink import SinksProviderMixin
from orbitalis.core.state import CoreState
from orbitalis.events.discover import DiscoverMessage, DiscoverQuery
from orbitalis.events.offer import OfferMessage, OfferedOperation
from orbitalis.core.requirement import Constraint, OperationRequirement
from orbitalis.events.reply import RequestOperationMessage, RejectOperationMessage
from orbitalis.events.response import ConfirmConnectionMessage, OperationNoLongerAvailableMessage
from orbitalis.orbiter.connection import Connection
from orbitalis.orbiter.orbiter import Orbiter
from orbitalis.orbiter.pending_request import PendingRequest
from orbitalis.orbiter.schemaspec import Input
from orbitalis.state_machine.state_machine import StateMachine


@dataclass(kw_only=True)
class Core(SinksProviderMixin, StateMachine[CoreState], Orbiter):
    """
    Component which connects itself to plugins, in order to be able to execute their operations.

    Author: Nicola Ricciardi
    """

    discovering_interval: float = field(default=2)
    operation_requirements: Dict[str, OperationRequirement] = field(default_factory=dict)    # operation_name => OperationRequirement

    compliant_event: asyncio.Event = field(default_factory=asyncio.Event, init=False)
    not_compliant_event: asyncio.Event = field(default_factory=asyncio.Event, init=False)

    _last_discover_sent_at: Optional[datetime] = field(default=None)

    def __post_init__(self):
        super().__post_init__()

        self.state = CoreState.CREATED
        self.compliant_event.clear()
        self.not_compliant_event.set()

    @classmethod
    def _is_plugin_operation_required_by_constraint(cls, constraint: Constraint) -> bool:
        return constraint.minimum > 0 \
            or (constraint.mandatory is not None and len(constraint.mandatory) > 0) \
            or constraint.maximum is None \
            or (constraint.maximum is not None and constraint.maximum > 0)

    @property
    def offer_topic(self) -> str:
        return f"$handshake.{self.identifier}.offer"

    @property
    def response_topic(self) -> str:
        return f"$handshake.{self.identifier}.response"


    @override
    async def _internal_start(self, *args, **kwargs):
        await super()._internal_start(*args, **kwargs)

        await self.eventbus_client.subscribe(
                topic=self.response_topic,
                handler=self.__response_event_handler
            )

        # Offer subscription MUST be the second,
        # in order to have response event handler when offers will be able to be managed
        await self.eventbus_client.subscribe(
                topic=self.offer_topic,
                handler=self.__offer_event_handler
            )

        self.update_compliant()

        await self.send_discover_based_on_requirements()

    @override
    async def _internal_stop(self, *args, **kwargs):
        await super()._internal_stop(*args, **kwargs)

    @override
    async def _on_stopped(self, *args, **kwargs):
        await super()._on_stopped(*args, **kwargs)

        self.state = CoreState.STOPPED

    def _on_compliant(self):
        """
        Hook called when core becomes compliant
        """

    def _on_not_compliant(self):
        """
        Hook called when core becomes not compliant
        """

    @override
    async def _on_close_connection(self, connection: Connection):

        if connection.has_output:
            await self.eventbus_client.unsubscribe(connection.output_topic)

        self.update_compliant()

    def current_constraint_for_operation(self, operation_name: str) -> Constraint:
        """
        Return current constraint for operation based on current connections
        """

        if operation_name not in self.operation_requirements.keys():
            raise KeyError(f"operation {operation_name} is not required")

        constraint = copy.deepcopy(self.operation_requirements[operation_name]).constraint

        for connection in self.retrieve_connections(operation_name=operation_name):

            if constraint.mandatory is not None and connection.remote_identifier in constraint.mandatory:
                constraint.mandatory.remove(connection.remote_identifier)

            if constraint.maximum is not None:
                constraint.maximum = max(0, constraint.maximum - 1)

            constraint.minimum = max(0, constraint.minimum - 1)

        return constraint

    def is_compliant_for_operation(self, operation_name: str) -> bool:
        """
        Evaluate at run-time if core is compliant for given operation based on its configuration. It may be a time-consuming operation
        """

        constraint = self.current_constraint_for_operation(operation_name)

        if constraint.mandatory is not None and len(constraint.mandatory) > 0:
            return False

        if constraint.minimum > 0:
            return False

        return True

    def is_compliant(self) -> bool:
        """
        Evaluate at run-time if core is global compliant based on its configuration. It may be a very time-consuming operation
        """

        for operation_name in self.operation_requirements.keys():
            if not self.is_compliant_for_operation(operation_name):
                return False

        return True
    
    def switch_to_compliant(self):
        """
        Force core to compliant state
        """

        self.state = CoreState.COMPLIANT
        self._on_compliant()
        self.compliant_event.set()
        self.not_compliant_event.clear()

    def switch_to_not_compliant(self):
        """
        Force core to not compliant state
        """

        self.state = CoreState.NOT_COMPLIANT
        self._on_not_compliant()
        self.compliant_event.clear()
        self.not_compliant_event.set()

    def reset_compliance_events(self):
        self.compliant_event.clear()
        self.not_compliant_event.clear()

    def update_compliant(self):
        """
        Use `is_compliant` to update core's state
        """

        if self.state == CoreState.CREATED:

            if self.is_compliant():
                self.switch_to_compliant()
                return

            else:
                self.switch_to_not_compliant()
                return

        before_was_compliance = self.state == CoreState.COMPLIANT
        now_is_compliance = self.is_compliant()

        if before_was_compliance and not now_is_compliance:
            self.switch_to_not_compliant()

        elif not before_was_compliance and now_is_compliance:
            self.switch_to_compliant()

    def _operation_to_discover(self) -> Dict[str, Constraint]:
        """
        Return a dictionary `operation_name` => `constraint`, based on current connections. This operations should be discover
        """

        operation_requirements: Dict[str, Constraint] = {}
        for operation_name in self.operation_requirements.keys():
            constraint = self.current_constraint_for_operation(operation_name)

            discover_operation = Core._is_plugin_operation_required_by_constraint(constraint)

            if not discover_operation:
                logging.debug("%s: operation %s already satisfied", self, operation_name)
                continue

            operation_requirements[operation_name] = constraint

        return operation_requirements

    async def _on_send_discover(self, discover_message: DiscoverMessage):
        """
        Hook called before discover message is sent
        """

    async def send_discover_for_operations(self, operation_requirements: Dict[str, Constraint]):

        if len(operation_requirements) == 0:
            return

        discover_message = DiscoverMessage(
            core_identifier=self.identifier,
            queries=dict([(operation_name, DiscoverQuery.from_constraint(operation_name, constraint)) for operation_name, constraint in operation_requirements.items()]),
            offer_topic=self.offer_topic,
            core_keepalive_topic=self.keepalive_topic,
            core_keepalive_request_topic=self.keepalive_request_topic,
            considered_dead_after=self.consider_others_dead_after
        )

        await self._on_send_discover(discover_message)

        await self.eventbus_client.publish(
            self.discover_topic,
            discover_message
        )

        self._last_discover_sent_at = datetime.now()

    async def send_discover_based_on_requirements(self):
        operation_requirements: Dict[str, Constraint] = self._operation_to_discover()

        if len(operation_requirements) > 0:
            await self.send_discover_for_operations(operation_requirements)

    def _is_plugin_operation_required_and_pluggable(self, plugin_identifier: str, offered_operation: OfferedOperation) -> bool:

        not_satisfied_requirement = self.current_constraint_for_operation(offered_operation.name)

        if not Core._is_plugin_operation_required_by_constraint(not_satisfied_requirement):
            return False

        if not not_satisfied_requirement.is_compatible(plugin_identifier):
            return False

        if not not_satisfied_requirement.input_is_compatible(offered_operation.input):
            return False

        if not not_satisfied_requirement.output_is_compatible(offered_operation.output):
            return False

        return True

    def _build_operation_output_topic(self, plugin_identifier: str, operation_name: str) -> str:
        return f"{operation_name}.{self.identifier}.{plugin_identifier}.output"

    async def _get_setup_data(self, plugin_identifier: str, offered_operation: OfferedOperation) -> Optional[bytes]:
        """
        Hook called to obtain setup data which generally will be sent to plugins. By default, `default_setup_data` is used
        """

        return self.operation_requirements[offered_operation.name].default_setup_data

    async def __request_operation(self, plugin_identifier: str, reply_topic: str, offered_operation: OfferedOperation):

        logging.debug("%s: operations to request: %s", self, offered_operation)

        if not self.operation_requirements[offered_operation.name].constraint.output_is_compatible(offered_operation.output):
            return  # invalid offer

        output_topic: Optional[str] = None

        if offered_operation.output.has_output:
            output_topic = self._build_operation_output_topic(plugin_identifier, offered_operation.name)


        setup_data = await self._get_setup_data(plugin_identifier, offered_operation)

        incoming_close_connection_topic: str = self._build_incoming_close_connection_topic(
            plugin_identifier,
            offered_operation.name
        )

        self._add_pending_request(PendingRequest(
            operation_name=offered_operation.name,
            remote_identifier=plugin_identifier,
            output_topic=output_topic,
            output=offered_operation.output,
            input=offered_operation.input,
            incoming_close_connection_topic=incoming_close_connection_topic
        ))

        await self.eventbus_client.publish(
            reply_topic,
            RequestOperationMessage(
                core_identifier=self.identifier,
                operation_name=offered_operation.name,
                response_topic=self.response_topic,
                output_topic=output_topic,
                core_side_close_operation_connection_topic=incoming_close_connection_topic,
                setup_data=setup_data
            )
        )


    async def __reject_operation(self, reply_topic: str, offered_operation: OfferedOperation):
        await self.eventbus_client.publish(
            reply_topic,
            RejectOperationMessage(
                core_identifier=self.identifier,
                operation_name=offered_operation.name
            )
        )

    async def _on_new_offer(self, offer_message: OfferMessage):
        """
        Hook called when a new offer arrives
        """

    @event_handler
    async def __offer_event_handler(self, topic: str, event: Event[OfferMessage]):
        logging.info("%s: new offer: %s -> %s", self, topic, event)

        await self._on_new_offer(event.payload)

        self.update_acquaintances(
            event.payload.plugin_identifier,
            keepalive_topic=event.payload.plugin_keepalive_topic,
            keepalive_request_topic=event.payload.plugin_keepalive_request_topic,
            consider_me_dead_after=event.payload.considered_dead_after
        )

        self.have_seen(event.payload.plugin_identifier)

        self._others_considers_me_dead_after[event.payload.plugin_identifier] = event.payload.considered_dead_after

        tasks = []
        for offered_operation in event.payload.offered_operations:
            if self._is_plugin_operation_required_and_pluggable(event.payload.plugin_identifier, offered_operation):
                tasks.append(asyncio.create_task(self.__request_operation(
                    event.payload.plugin_identifier,
                    event.payload.reply_topic,
                    offered_operation
                )))
            else:
                tasks.append(asyncio.create_task(self.__reject_operation(
                    event.payload.reply_topic,
                    offered_operation
                )))

        try:
            await asyncio.gather(*tasks)

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            if self.raise_exceptions:
                raise e

    async def _on_confirm_connection(self, confirm_connection_message: ConfirmConnectionMessage):
        """
        Hook called when a confirm connection arrives
        """

    async def __confirm_connection_event_handler(self, topic: str, event: Event[ConfirmConnectionMessage]):
        await self._on_confirm_connection(event.payload)

        plugin_identifier = event.payload.plugin_identifier
        operation_name = event.payload.operation_name

        if not self._is_pending(plugin_identifier, operation_name):
            logging.warning("%s: operation %s from plugin %s was not requested", self, operation_name, plugin_identifier)
            return

        pending_request = self._pending_requests_by_remote_identifier(plugin_identifier)[operation_name]

        async with pending_request.lock:
            if not self._is_pending(plugin_identifier, operation_name):
                logging.warning("%s: pending request (%s, %s) not available anymore", self, plugin_identifier, operation_name)
                return

            topics_to_unsubscribe_if_error: List[str] = []
            try:

                await self.eventbus_client.subscribe(
                    pending_request.incoming_close_connection_topic,
                    self._close_connection_event_handler
                )

                topics_to_unsubscribe_if_error.append(pending_request.incoming_close_connection_topic)

            except Exception as e:
                logging.error("%s: error during subscribing on close connection in response handling: %s", self, repr(e))

                if self.raise_exceptions:
                    raise e

            if pending_request.output_topic is not None:        # output is excepted
                handler: Optional[EventHandler] = None
                if operation_name in self.operation_sinks:
                    handler = self.operation_sinks[operation_name]

                if self.operation_requirements[operation_name].has_override_sink:
                    handler = self.operation_requirements[operation_name].override_sink

                if handler is not None:
                    try:
                        await self.eventbus_client.subscribe(
                            pending_request.output_topic,
                            handler
                        )

                        topics_to_unsubscribe_if_error.append(pending_request.output_topic)

                    except Exception as e:
                        logging.error("%s: error during subscribing to '%s' in response handling: %s", self, pending_request.output_topic, repr(e))

                        await self.eventbus_client.unsubscribe(pending_request.incoming_close_connection_topic)

                        if self.raise_exceptions:
                            raise e

            pending_request.input_topic = event.payload.operation_input_topic
            pending_request.close_connection_to_remote_topic = event.payload.plugin_side_close_operation_connection_topic

            try:
                self._promote_pending_request_to_connection(pending_request)

            except Exception as e:
                logging.error("%s: error during pending request promoting in response handling: %s", self, repr(e))

                await self.eventbus_client.multi_unsubscribe(topics_to_unsubscribe_if_error, parallelize=True)

                if self.raise_exceptions:
                    raise e

    async def _on_operation_no_longer_available(self, message: OperationNoLongerAvailableMessage):
        """
        Hook called when operation no longer available message arrives
        """

    async def __operation_no_longer_available_event_handler(self, topic: str, event: Event[OperationNoLongerAvailableMessage]):
        await self._on_operation_no_longer_available(event.payload)

        try:
            pending_request = self._pending_requests[event.payload.plugin_identifier][event.payload.operation_name]

            async with pending_request.lock:
                self._remove_pending_request(pending_request)

        except Exception as e:
            logging.warning("%s: pending request for operation %s of plugin %s can not be removed, maybe already removed", self, event.payload.operation_name, event.payload.plugin_identifier)

    async def _on_response(self):
        """
        Hook called when response message arrives
        """

    @event_handler
    async def __response_event_handler(self, topic: str, event: Event[ConfirmConnectionMessage | OperationNoLongerAvailableMessage]):

        logging.info("%s: new response: %s -> %s", self, topic, event)

        self.have_seen(event.payload.plugin_identifier)

        await self._on_response()

        if isinstance(event.payload, ConfirmConnectionMessage):
            await self.__confirm_connection_event_handler(topic, event)

        elif isinstance(event.payload, OperationNoLongerAvailableMessage):
            await self.__operation_no_longer_available_event_handler(topic, event)

        else:
            raise ValueError("Unexpected reply message")

        self.update_compliant()

    async def execute_distributed(self, operation_name: str, data: List[Optional[AvroMessageMixin]], fire_and_forget: bool = False) -> Set[str]:
        """
        Execute the operation by its name, distributing provided data among all compatible plugins.
        All data messages must be of the same type.

        Return the plugin identifiers the data has been sent to.
        """

        if len(data) == 0:
            return set()
        
        message_type: Type[AvroMessageMixin] | Type[None] = type(data[0])
        for index in range(1, len(data)):
            if not isinstance(data[index], message_type):
                raise ValueError("all data messages must be of the same type")

        connections = self.retrieve_connections(
            operation_name=operation_name,
            input=Input.empty() if message_type is type(None) else Input.from_message(message_type)
        )

        plugin_identifiers: Set[str] = set()
        publishes = []
        connection_index = 0

        for message in data:
            connection = connections[connection_index % len(connections)]
            connection_index += 1

            publishes.append(
                asyncio.create_task(
                    self.eventbus_client.publish(
                        connection.input_topic,
                        message
                    )
                )
            )

            plugin_identifiers.add(connection.remote_identifier)

        if not fire_and_forget:
            await asyncio.gather(*publishes)

        return plugin_identifiers

    async def execute_sending_all(self, operation_name: str, data: Optional[AvroMessageMixin] = None, fire_and_forget: bool = False) -> Set[str]:
        """
        Execute the operation by its name, sending provided data to all compatible plugins.

        Return the plugin identifiers the data has been sent to.
        """

        connections = self.retrieve_connections(
            operation_name=operation_name,
            input=Input.empty() if data is None else Input.from_message(type(data))
        )

        plugin_identifiers: Set[str] = set()
        publishes = []
        for connection in connections:
            plugin_identifiers.add(connection.remote_identifier)
            
            publishes.append(
                asyncio.create_task(
                    self.eventbus_client.publish(
                        connection.input_topic,
                        data
                    )
                )
            )

        if not fire_and_forget:
            await asyncio.gather(*publishes)

        return plugin_identifiers
    
    async def execute_sending_any(self, operation_name: str, data: Optional[AvroMessageMixin] = None, fire_and_forget: bool = False) -> str:
        """
        Execute the operation by its name, sending provided data to only one random compatible plugin.

        Return the plugin identifier the data has been sent to.
        """

        connections = self.retrieve_connections(
            operation_name=operation_name,
            input=Input.empty() if data is None else Input.from_message(type(data))
        )

        connection = random.choice(connections)

        task = asyncio.create_task(
            self.eventbus_client.publish(
                connection.input_topic,
                data
            )
        )

        if not fire_and_forget:
            await task

        return connection.remote_identifier

    async def execute_using_plugin(self, operation_name: str, plugin_identifier: str, data: Optional[AvroMessageMixin] = None, fire_and_forget: bool = False):
        """
        Execute the operation by its name, sending provided data to only one specific plugin.
        """

        connections = self.retrieve_connections(
            operation_name=operation_name,
            remote_identifier=plugin_identifier,
            input=Input.empty() if data is None else Input.from_message(type(data))
        )

        if len(connections) == 0:
            raise ValueError(f"no connection found for operation {operation_name} and plugin {plugin_identifier}")

        if len(connections) > 1:
            raise ValueError(f"multiple connections found for operation {operation_name} and plugin {plugin_identifier}")
        
        connection = connections[0]

        if connection.remote_identifier != plugin_identifier:
            raise ValueError(f"connection found for operation {operation_name} does not match plugin {plugin_identifier}")

        task = asyncio.create_task(
            self.eventbus_client.publish(
                connection.input_topic,
                data
            )
        )

        if not fire_and_forget:
            await task

    async def execute(self, operation_name: str, data: Optional[AvroMessageMixin] | List[Optional[AvroMessageMixin]] = None, fire_and_forget: bool = False,
                      *, any: Optional[bool] = None, all: Optional[bool] = None, plugin_identifier: Optional[str] = None, distribute: Optional[bool] = None) -> Set[str]:
        """
        Execute the operation by its name, sending provided data.

        Return the plugin identifiers the data has been sent to.
        """

        if any is None and all is None and plugin_identifier is None and distribute is None:
            raise ValueError("mode (any/all/identifier/distribute) must be specified")

        messages: List[Optional[AvroMessageMixin]] = data if isinstance(data, list) else [data]

        if distribute is not None and distribute:
            return await self.execute_distributed(operation_name=operation_name, data=messages, fire_and_forget=fire_and_forget)

        if plugin_identifier is not None:
            for message in messages:
                sent = await self.execute_using_plugin(operation_name=operation_name, data=message, plugin_identifier=plugin_identifier, fire_and_forget=fire_and_forget)
                if not sent:
                    raise ValueError(f"can not send data to plugin {plugin_identifier} for operation {operation_name}")

            return {plugin_identifier}
        
        if any is not None and any:
            plugin_identifiers: Set[str] = set()
            for message in messages:
                plugin_id = await self.execute_sending_any(operation_name=operation_name, data=message, fire_and_forget=fire_and_forget)
                plugin_identifiers.add(plugin_id)

            return plugin_identifiers
        
        if all is not None and all:
            plugin_identifiers: Set[str] = set()
            for message in messages:
                sent_plugin_ids = await self.execute_sending_all(operation_name=operation_name, data=message, fire_and_forget=fire_and_forget)
                plugin_identifiers.update(sent_plugin_ids)

            return plugin_identifiers
        
        raise ValueError("invalid mode specified")

    async def sudo_execute(self, topic: str, data: Optional[AvroMessageMixin] = None):
        """
        Bypass all checks and send data to topic
        """

        await self.eventbus_client.publish(topic, data)

    @override
    async def _on_loop_start(self):
        await asyncio.sleep(self.discovering_interval)

    @override
    async def _on_loop_iteration(self):
        self.update_compliant()

        if self._last_discover_sent_at is None or (self._last_discover_sent_at + timedelta(seconds=self.discovering_interval)) < datetime.now():
            await self.send_discover_based_on_requirements()

    def __str__(self):
        return f"Core('{self.identifier}')"





from dataclasses import dataclass, field
from typing import Optional

from busline.event.registry import add_to_registry
from busline.event.message.avro_message import AvroMessageMixin


@dataclass(frozen=True, kw_only=True)
class RequestOperationMessage(AvroMessageMixin):
    """
    Core --- request ---> Plugin

    Message used by core to formally request an operation. Every operation has own request.
    Core provides additional information to finalize the connection

    Author: Nicola Ricciardi
    """

    core_identifier: str
    operation_name: str
    response_topic: str
    output_topic: Optional[str]
    core_side_close_operation_connection_topic: str
    setup_data: Optional[bytes]


@dataclass(frozen=True, kw_only=True)
class RejectOperationMessage(AvroMessageMixin):
    """
    Core --- reject ---> Plugin

    Message used by core to formally reject an operation (e.g., not needed anymore). Every operation has own reject.

    Author: Nicola Ricciardi
    """

    core_identifier: str
    operation_name: str
    description: Optional[str] = field(default=None)



from dataclasses import dataclass, field
from typing import Optional, List

from busline.client.subscriber.event_handler.event_handler import EventHandler
from orbitalis.orbiter.schemaspec import Inputs, Outputs
from orbitalis.utils.allowblocklist import AllowBlockListMixin


@dataclass(kw_only=True)
class Constraint(AllowBlockListMixin, Inputs, Outputs):
    """
    Constraint usable for needs

    min: minimum number (mandatory included)
    max: maximum number (mandatory included)
    mandatory: identifiers

    Author: Nicola Ricciardi
    """

    minimum: int = field(default=0)
    maximum: Optional[int] = field(default=None)
    mandatory: List[str] = field(default_factory=list)


    def __post_init__(self):
        super().__post_init__()

        if self.minimum < 0 or (self.maximum is not None and self.maximum < 0) \
                or (self.maximum is not None and self.minimum > self.maximum):
            raise ValueError("Minimum and/or maximum value are invalid")

        if len(self.inputs) == 0:
            raise ValueError("Missed inputs")

        if len(self.outputs) == 0:
            raise ValueError("Missed outputs")


@dataclass
class OperationRequirement:
    """
    Operation requirement for a Core to be compliant

    Author: Nicola Ricciardi
    """

    constraint: Constraint
    override_sink: Optional[EventHandler] = field(default=None, kw_only=True)
    default_setup_data: Optional[bytes] = field(default=None, kw_only=True)

    @property
    def has_override_sink(self) -> bool:
        return self.override_sink is not None



from __future__ import annotations

import asyncio
import logging
from typing import override, List, Optional, Any
from dataclasses import dataclass

from uuid import uuid4

from busline.client.subscriber.event_handler import event_handler
from busline.event.event import Event
from orbitalis.core.requirement import Constraint
from orbitalis.events.discover import DiscoverMessage, DiscoverQuery
from orbitalis.events.offer import OfferMessage, OfferedOperation
from orbitalis.events.reply import RejectOperationMessage, RequestOperationMessage
from orbitalis.events.response import ConfirmConnectionMessage, OperationNoLongerAvailableMessage
from orbitalis.orbiter.connection import Connection
from orbitalis.orbiter.orbiter import Orbiter
from orbitalis.orbiter.pending_request import PendingRequest
from orbitalis.plugin.operation import OperationsProviderMixin
from orbitalis.plugin.state import PluginState
from orbitalis.state_machine.state_machine import StateMachine


@dataclass(kw_only=True)
class Plugin(OperationsProviderMixin, StateMachine, Orbiter):
    """
    Component which provides a set of operations.

    Author: Nicola Ricciardi
    """

    def __post_init__(self):
        super().__post_init__()

        self.state = PluginState.CREATED

    @property
    def reply_topic(self) -> str:
        return f"$handshake.{self.identifier}.reply"

    @override
    async def _internal_start(self, *args, **kwargs):
        await super()._internal_start(*args, **kwargs)

        await self.eventbus_client.subscribe(
            self.reply_topic,
            self.__reply_event_handler
        )

        await self.eventbus_client.subscribe(
            self.discover_topic,
            self.__discover_event_handler
        )

        self.state = PluginState.RUNNING

    @override
    async def _internal_stop(self, *args, **kwargs):
        await super()._internal_stop(*args, **kwargs)

        topics: List[str] = [
            self.discover_topic,
            self.reply_topic
        ]

        await self.eventbus_client.multi_unsubscribe(topics, parallelize=True)

    @override
    async def _on_stopped(self, *args, **kwargs):
        await super()._on_stopped(*args, **kwargs)

        self.state = PluginState.STOPPED

    @override
    async def _on_close_connection(self, connection: Connection):

        if connection.has_input:
            await self.eventbus_client.unsubscribe(connection.input_topic)

    def __can_lend_to_core(self, core_identifier: str, operation_name: str) -> bool:
        if not self.operations[operation_name].policy.is_compatible(core_identifier):
            return False

        if self.operations[operation_name].policy.maximum is None or len(self.retrieve_connections(operation_name=operation_name)) < self.operations[operation_name].policy.maximum:
            return True

        return False

    def __allow_offer(self, core_identifier: str, core_needed_operation_name: str, core_discover_query: DiscoverQuery) -> bool:
        if core_needed_operation_name not in self.operations:
            return False

        # check compatibility with block/allow list
        if not core_discover_query.is_compatible(self.identifier):
            return False

        # check if already in pending request
        if core_identifier in self._pending_requests.keys() \
                and core_needed_operation_name in self._pending_requests_by_remote_identifier(core_identifier):
            return False

        # check if this plugin have already lent operation to core
        if len(self.retrieve_connections(remote_identifier=core_identifier,
                                         operation_name=core_needed_operation_name)) > 0:
            return False

        # check if there are slot available
        if self.operations[core_needed_operation_name].policy.maximum is not None:
            current_reserved_slot_for_operation: int = len(
                self.retrieve_connections(operation_name=core_needed_operation_name))

            for core_identifier, operations in self._pending_requests.items():
                if core_needed_operation_name in operations.keys():
                    current_reserved_slot_for_operation += 1

            if current_reserved_slot_for_operation >= self.operations[core_needed_operation_name].policy.maximum:
                return False

        # check input_schemas compatibility
        if not core_discover_query.input_is_compatible(self.operations[core_needed_operation_name].input):
            return False

        # check output_schemas compatibility
        if not core_discover_query.output_is_compatible(self.operations[core_needed_operation_name].output):
            return False

        if not self.__can_lend_to_core(core_identifier, core_needed_operation_name):
            return False

        return True

    async def _on_new_discover(self, discover_message: DiscoverMessage):
        """
        Hook called when a new discover message arrives
        """

    @event_handler
    async def __discover_event_handler(self, topic: str, event: Event[DiscoverMessage]):
        logging.info("%s: new discover event from %s: %s -> %s", self, event.payload.core_identifier, topic, event)

        await self._on_new_discover(event.payload)

        self.update_acquaintances(
            event.payload.core_identifier,
            keepalive_topic=event.payload.core_keepalive_topic,
            keepalive_request_topic=event.payload.core_keepalive_request_topic,
            consider_me_dead_after=event.payload.considered_dead_after
        )

        self.have_seen(event.payload.core_identifier)

        self._others_considers_me_dead_after[event.payload.core_identifier] = event.payload.considered_dead_after

        offerable_operations: List[str] = []

        for core_needed_operation_name, core_needed_operation_constraint in event.payload.queries.items():

            if self.__allow_offer(event.payload.core_identifier, core_needed_operation_name, core_needed_operation_constraint):
                offerable_operations.append(core_needed_operation_name)

        logging.debug("%s: send offer for these operations: %s", self, offerable_operations)
        if len(offerable_operations) > 0:
            await self.send_offer(
                event.payload.offer_topic,
                event.payload.core_identifier,
                offerable_operations
            )


    def _build_operation_input_topic_for_core(self, core_identifier: str, operation_name: str) -> str:
        return f"{operation_name}.{core_identifier}.{self.identifier}.input.{uuid4()}"

    async def _on_send_offer(self, offer_message: OfferMessage):
        """
        Hook called before offer is sent
        """

    async def send_offer(self, offer_topic: str, core_identifier: str, offerable_operations: List[str]):
        """
        Send a new offer message in given topic to given core identifier
        """

        if len(offerable_operations) == 0:
            return

        offered_operations: List[OfferedOperation] = []
        new_pending_requests: List[PendingRequest] = []

        for operation_name in offerable_operations:
            offered_operations.append(
                OfferedOperation(
                    name=operation_name,
                    input=self.operations[operation_name].input,
                    output=self.operations[operation_name].output
                )
            )

            pending_request = PendingRequest(
                operation_name=operation_name,
                remote_identifier=core_identifier,
                input=self.operations[operation_name].input,
                output=self.operations[operation_name].output
            )

            self._add_pending_request(pending_request)
            new_pending_requests.append(pending_request)

        try:
            offer_message = OfferMessage(
                    plugin_identifier=self.identifier,
                    offered_operations=offered_operations,
                    reply_topic=self.reply_topic,
                    plugin_keepalive_topic=self.keepalive_topic,
                    plugin_keepalive_request_topic=self.keepalive_request_topic,
                    considered_dead_after=self.consider_others_dead_after
                )

            await self._on_send_offer(offer_message)

            await self.eventbus_client.publish(offer_topic, offer_message)

        except Exception as e:
            logging.error("%s: %s", self, repr(e))

            for pending_request in new_pending_requests:
                try:
                    async with pending_request.lock:
                        self._remove_pending_request(pending_request)
                except Exception:
                    pass

            if self.raise_exceptions:
                raise e

    async def _on_reject(self, message: RejectOperationMessage):
        """
        Hook called when a reject message arrives
        """

    async def __reject_event_handler(self, topic: str, event: Event[RejectOperationMessage]):
        logging.debug("%s: core %s rejects plug request for this operation: %s", self, event.payload.core_identifier, event.payload.operation_name)

        await self._on_reject(event.payload)

        try:
            pending_request = self._pending_requests[event.payload.core_identifier][event.payload.operation_name]

            async with pending_request.lock:
                self._remove_pending_request(pending_request)

        except Exception as e:
            logging.warning("%s: pending request ('%s', '%s') can not be removed", self, event.payload.core_identifier, event.payload.operation_name)

    async def _setup_operation(self, core_identifier: str, operation_name: str, setup_data: Optional[bytes]):
        """
        Hook called to set up operation when connection is created
        """

    async def _plug_operation_into_core(self, core_identifier: str, response_topic: str, operation_name: str, setup_data: Optional[bytes]):
        """

        Return (operation_input_topic, plugin_side_close_operation_connection_topic)
        """

        topics_to_unsubscribe_if_error: List[str] = []

        operation_input_topic: str = self._build_operation_input_topic_for_core(core_identifier, operation_name)

        plugin_side_close_operation_connection_topic = self._build_incoming_close_connection_topic(
            core_identifier,
            operation_name
        )

        try:
            await self.eventbus_client.subscribe(operation_input_topic, self.operations[operation_name].handler)
            topics_to_unsubscribe_if_error.append(operation_input_topic)

            await self.eventbus_client.subscribe(
                plugin_side_close_operation_connection_topic,
                self._close_connection_event_handler
            )
            topics_to_unsubscribe_if_error.append(plugin_side_close_operation_connection_topic)

            if setup_data is not None:
                await self._setup_operation(
                    core_identifier,
                    operation_name,
                    setup_data
                )

            await self.eventbus_client.publish(
                response_topic,
                ConfirmConnectionMessage(
                    plugin_identifier=self.identifier,
                    operation_name=operation_name,
                    operation_input_topic=operation_input_topic,
                    plugin_side_close_operation_connection_topic=plugin_side_close_operation_connection_topic
                )
            )

            return operation_input_topic, plugin_side_close_operation_connection_topic

        except Exception as e:
            logging.error("%s: error during plug operation '%s' into core '%s': %s", self, operation_name, core_identifier, repr(e))

            await self.eventbus_client.multi_unsubscribe(topics_to_unsubscribe_if_error, parallelize=True)

            raise e

    async def _on_request(self, message: RequestOperationMessage):
        """
        Hook called when a new request message arrives
        """

    async def __request_operation_event_handler(self, topic: str, event: Event[RequestOperationMessage]):

        await self._on_request(event.payload)

        core_identifier = event.payload.core_identifier
        operation_name = event.payload.operation_name

        logging.debug("%s: core %s confirms plug request for this operation: %s", self, core_identifier, operation_name)

        if not self._is_pending(core_identifier, operation_name):
            logging.warning("%s: pending request for ('%s', '%s') not found", self, core_identifier, operation_name)
            return

        pending_request = self._pending_requests_by_remote_identifier(core_identifier)[operation_name]

        async with pending_request.lock:
            if not self._is_pending(core_identifier, operation_name):
                logging.warning("%s: pending request (%s, %s) not available anymore", self, core_identifier, operation_name)
                return

            try:
                if not self.__can_lend_to_core(core_identifier, operation_name):
                    logging.debug("%s: can not lend to core '%s' operation: %s", self, core_identifier, operation_name)

                    await self.eventbus_client.publish(
                        event.payload.response_topic,
                        OperationNoLongerAvailableMessage(
                            plugin_identifier=self.identifier,
                            operation_name=operation_name
                        )
                    )

                    self._remove_pending_request(pending_request)

                else:

                    operation_input_topic, plugin_side_close_operation_connection_topic = await self._plug_operation_into_core(
                        core_identifier,
                        event.payload.response_topic,
                        operation_name,
                        event.payload.setup_data
                    )

                    pending_request.incoming_close_connection_topic = plugin_side_close_operation_connection_topic
                    pending_request.input_topic = operation_input_topic
                    pending_request.output_topic = event.payload.output_topic
                    pending_request.close_connection_to_remote_topic = event.payload.core_side_close_operation_connection_topic

                    self._promote_pending_request_to_connection(pending_request)

            except Exception as e:
                logging.error("%s: error during confirm pending request': %s", self, repr(e))

                if self.raise_exceptions:
                    raise e


    async def _on_reply(self):
        """
        Hook called when a new reply message arrives
        """

    @event_handler
    async def __reply_event_handler(self, topic: str, event: Event[RequestOperationMessage | RejectOperationMessage]):
        logging.info("%s: new reply: %s -> %s", self, topic, event)

        self.have_seen(event.payload.core_identifier)

        await self._on_reply()

        if isinstance(event.payload, RequestOperationMessage):
            await self.__request_operation_event_handler(topic, event)

        elif isinstance(event.payload, RejectOperationMessage):
            await self.__reject_event_handler(topic, event)

        else:
            raise ValueError("Unexpected reply message")

    async def send_result_to_all(self, connections: List[Connection], data):
        """
        Send data to all connections which have an output (checking `connection.has_output`)
        """

        tasks = []
        for connection in connections:

            # Only if the connection expects an output
            # it is published in the related topic
            if connection.has_output:
                tasks.append(
                    asyncio.create_task(
                        self.eventbus_client.publish(
                            connection.output_topic,
                            data
                        )
                    )
                )

        await asyncio.gather(*tasks)  # wait publishes

    def __str__(self):
        return f"Plugin('{self.identifier}')"





import asyncio
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List
from orbitalis.orbiter.connection import Connection
from orbitalis.orbiter.schemaspec import SchemaSpec, Input, Output



@dataclass
class PendingRequest:
    operation_name: str
    remote_identifier: str

    lock: asyncio.Lock = field(default_factory=lambda: asyncio.Lock(), init=False)

    input: Optional[Input] = field(default=None)
    output: Optional[Output] = field(default=None)

    input_topic: Optional[str] = field(default=None, kw_only=True)
    incoming_close_connection_topic: Optional[str] = field(default=None, kw_only=True)
    close_connection_to_remote_topic: Optional[str] = field(default=None, kw_only=True)
    output_topic: Optional[str] = field(default=None, kw_only=True)

    created_at: datetime = field(default_factory=lambda: datetime.now(), init=False)

    def into_connection(self) -> Connection:

        if self.input_topic is None:
            raise ValueError("input_topic missed")

        if self.input is None:
            raise ValueError("input missed")

        if self.incoming_close_connection_topic is None:
            raise ValueError("incoming_close_connection_topic missed")

        if self.close_connection_to_remote_topic is None:
            raise ValueError("close_connection_to_remote_topic missed")

        if self.output is None:
            raise ValueError("output missed")


        return Connection(
            operation_name=self.operation_name,
            remote_identifier=self.remote_identifier,
            input_topic=self.input_topic,
            input=self.input,
            output_topic=self.output_topic,
            output=self.output,
            incoming_close_connection_topic=self.incoming_close_connection_topic,
            close_connection_to_remote_topic=self.close_connection_to_remote_topic,
        )



from dataclasses import dataclass
from typing import Dict

from dataclasses_avroschema import AvroModel

from busline.event.registry import add_to_registry
from busline.event.message.avro_message import AvroMessageMixin
from orbitalis.core.requirement import Constraint
from orbitalis.orbiter.schemaspec import Outputs, Inputs
from orbitalis.utils.allowblocklist import AllowBlockListMixin


@dataclass
class DiscoverQuery(AllowBlockListMixin, Inputs, Outputs, AvroModel):

    operation_name: str

    @classmethod
    def from_constraint(cls, operation_name: str, constraint: Constraint):
        return cls(
            operation_name=operation_name,
            inputs=constraint.inputs,
            outputs=constraint.outputs,
            allowlist=constraint.allowlist,
            blocklist=constraint.blocklist,
        )



@dataclass(frozen=True)
class DiscoverMessage(AvroMessageMixin):
    """
    Core --- discover ---> Plugin

    Message used by cores to notify plugins of their presence and to ask operations connections,
    core provides a full set of pluggable operations with related information

    Author: Nicola Ricciardi
    """

    core_identifier: str
    offer_topic: str
    core_keepalive_topic: str
    core_keepalive_request_topic: str
    considered_dead_after: float
    queries: Dict[str, DiscoverQuery]   # operation_name => DiscoverQuery



from dataclasses import dataclass,field
from abc import ABC
from typing import Optional, Dict, Set, List, Self

from dataclasses_avroschema import AvroModel


@dataclass(kw_only=True)
class AllowBlockListMixin(AvroModel, ABC):
    """
    allowlist: admitted Orbs (by identifiers)
    blocklist: not admitted Orbs (by identifiers)
    priority: identifier => priority (int)

    Author: Nicola Ricciardi
    """

    allowlist: Optional[List[str]] = field(default=None)
    blocklist: Optional[List[str]] = field(default=None)

    def __post_init__(self):
        if self.allowlist is not None and self.blocklist is not None:
            raise ValueError("allowlist and blocklist can not be used together")

    @classmethod
    def allow_only(cls, identifier: str) -> Self:
        return cls(allowlist=[identifier])

    def is_compatible(self, identifier: str) -> bool:
        if self.blocklist is not None and identifier in self.blocklist:
            return False

        if self.allowlist is not None and identifier not in self.allowlist:
            return False

        return True




import asyncio
from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional, List, Self

from orbitalis.orbiter.schemaspec import Input, Output


@dataclass
class Connection:
    """
    Orbiter2 uses this connection to execute related operation on Orbiter1.

    Orbiter1 <--- input_topic --- Orbiter2
    Orbiter1 --- output_topic ---> Orbiter2     (optional, only if there is an output)

    Orbiter (you) --- close_connection_to_remote_topic ---> Orbiter (remote)
    Orbiter (you) <--- close_connection_to_local_topic --- Orbiter (remote)

    Author: Nicola Ricciardi
    """

    operation_name: str
    remote_identifier: str

    incoming_close_connection_topic: str
    close_connection_to_remote_topic: str

    lock: asyncio.Lock = field(default_factory=lambda: asyncio.Lock(), init=False)

    input: Input
    output: Output

    input_topic: Optional[str] = field(default=None)
    output_topic: Optional[str] = field(default=None)

    soft_closed_at: Optional[datetime] = field(default=None, init=False)
    created_at: datetime = field(default_factory=lambda: datetime.now())
    last_use: Optional[datetime] = field(default=None)

    @property
    def is_soft_closed(self) -> bool:
        return self.soft_closed_at is not None

    @property
    def has_input(self) -> bool:
        return self.input_topic is not None

    @property
    def has_output(self) -> bool:
        return self.output_topic is not None

    def touch(self):
        """
        Update last use
        """
        self.last_use = datetime.now()

    def soft_close(self):
        if self.soft_closed_at is None:
            self.soft_closed_at = datetime.now()

    def __str__(self):
        return f"('{self.remote_identifier}', '{self.operation_name}')"


import inspect
from abc import ABC
from dataclasses import dataclass, field
from typing import Optional, Dict, Self, Any
from busline.event.message.avro_message import AvroMessageMixin
from busline.client.subscriber.event_handler import event_handler
from busline.client.subscriber.event_handler.event_handler import EventHandler
from orbitalis.orbiter.schemaspec import Input, Output
from orbitalis.utils.allowblocklist import AllowBlockListMixin


@dataclass
class Policy(AllowBlockListMixin):
    maximum: Optional[int] = field(default=None)

    @classmethod
    def no_constraints(cls) -> Self:
        return cls(maximum=None, allowlist=None, blocklist=None)


@dataclass(kw_only=True)
class Operation:
    name: str
    handler: Optional[EventHandler]
    policy: Policy
    input: Input
    output: Output

    def __post_init__(self):
        if self.input.has_input and self.handler is None:
            raise ValueError("Missed handler")


@dataclass(kw_only=True)
class _OperationDescriptor:
    operation_name: str
    func: Any
    policy: Policy
    input: Input
    output: Output

    def __post_init__(self):
        self.func = event_handler(self.func)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        if self.operation_name not in instance.operations:
            instance.operations[self.operation_name] = Operation(
                name=self.operation_name,
                handler=self.func.__get__(instance, owner),
                policy=self.policy,
                input=self.input,
                output=self.output
            )

        return self.func.__get__(instance, owner)


def operation(*, input: Optional[Input | AvroMessageMixin] = None, default_policy: Optional[Policy] = None, output: Optional[Output | AvroMessageMixin] = None, name: Optional[str] = None):
    """
    Transform a function of a method in an operation and append it to operations provider
    """

    if input is None:
        input = Input.no_input()

    if isinstance(input, AvroMessageMixin):
        input = Input.from_message(type(input))

    if output is None:
        output = Output.no_output()

    if isinstance(output, AvroMessageMixin):
        output = Output.from_message(type(output))

    if default_policy is None:
        default_policy = Policy.no_constraints()

    def decorator(func):
        if not inspect.iscoroutinefunction(func):
            raise TypeError("Event handler must be async")

        op_name = name or func.__name__

        return _OperationDescriptor(
            func=func,
            operation_name=op_name,
            policy=default_policy,
            input=input,
            output=output
        )

    return decorator


@dataclass(kw_only=True)
class OperationsProviderMixin(ABC):

    operations: Dict[str, Operation] = field(default_factory=dict)     # operation_name => Operation


    def __post_init__(self):

        # used to refresh operations
        for attr_name in dir(self):
            _ = getattr(self, attr_name)

    def with_operation(self, operation_name: str, operation: Operation) -> Self:
        self.operations[operation_name] = operation

        return self

    def with_custom_policy(self, operation_name: str, policy: Policy) -> Self:
        self.operations[operation_name].policy = policy

        return self



from dataclasses import dataclass

from busline.event.message.avro_message import AvroMessageMixin
from busline.event.registry import add_to_registry


@dataclass(frozen=True, kw_only=True)
class KeepaliveRequestMessage(AvroMessageMixin):
    """
    Orbiter A --- keepalive_request ---> Orbiter B

    Message sent to request a keepalive

    Author: Nicola Ricciardi
    """

    from_identifier: str
    keepalive_topic: str


@dataclass(frozen=True, kw_only=True)
class KeepaliveMessage(AvroMessageMixin):
    """
    Orbiter A <--- keepalive --- Orbiter B

    Message sent to notify a keepalive

    Author: Nicola Ricciardi
    """

    from_identifier: str


from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
import logging
from enum import Enum
from typing import TypeVar, Generic



S = TypeVar('S', bound=Enum)


class StateMachine(Generic[S], ABC):
    __state: S = None

    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, s: S):
        logging.info("%s: %s --> %s", self, self.__state.name if self.__state is not None else 'None', s.name)
        self.__state = s



from dataclasses import dataclass
from typing import Optional, TypeVar, Generic

from dataclasses_avroschema import AvroModel

from busline.event.registry import add_to_registry
from busline.event.message.avro_message import AvroMessageMixin


@dataclass(frozen=True, kw_only=True)
class GracelessCloneConnectionMessage(AvroMessageMixin):
    """
    Orbiter A --- close ---> Orbiter B

    Message sent to close connection

    Author: Nicola Ricciardi
    """

    from_identifier: str
    operation_name: str
    data: Optional[bytes]


@dataclass(frozen=True, kw_only=True)
class GracefulCloseConnectionMessage(AvroMessageMixin):
    """
    Orbiter A --- close ---> Orbiter B

    Message sent to notify the desire to close the connection

    Author: Nicola Ricciardi
    """

    from_identifier: str
    operation_name: str
    ack_topic: str
    data: Optional[bytes]


@dataclass(frozen=True, kw_only=True)
class CloseConnectionAckMessage(AvroMessageMixin):
    """
    Orbiter A <--- close ack --- Orbiter B

    Message sent to confirm close connection

    Author: Nicola Ricciardi
    """

    from_identifier: str
    operation_name: str

